/***********************************************************************
 *  File: measure_decl.h
 *  Date: Torino, April 20 1993
 *  Programmer: Giovanni Chiola
 ***********************************************************************/

#include "simul_decl.h"
#include "../gsrc2/const.h"
#include "../gsrc2/grg.h"


#define MAX_TOKEN 32767


extern FILE * nfp, * mpdfp, * tpdfp;

struct Codbyt_descr {
  unsigned char shift;
  unsigned char mask;
  unsigned short num_byte;
  unsigned short link;
};

struct Place_descr {
    unsigned char lbound;
    unsigned char rubound;
    struct Codbyt_descr cod;
    unsigned short alb;
    unsigned short aub;
    double * tok_distr;
    double * cum_distr;
};
 
typedef struct Place_descr * Place_p;
 
struct Byte_descr {
  unsigned char noitems;
  unsigned char max_conf;
  unsigned short first_item;
};
 
struct p_bound_s {
    unsigned short min_t;
    unsigned short max_t;
};


struct Cond_descr {
  char op;
  char type;
  char top1;
  char top2;
  int op1;
  int op2;
};


struct Prob_descr {
  char type;
  int place;
  struct Cond_descr * cond;
  double val;
  double cumval;
};
typedef struct Prob_descr * Prob_p;

union Op_val {
  unsigned num;
  double rval;
};

struct Op_descr {
  char op;
  char type;
  union Op_val val;
};

struct Res_descr {
  char type;
  char * name;
  struct Op_descr * oper;
  double val;
  double cumval;
};
typedef struct Res_descr * Res_p;

struct Trans_descr {
    unsigned short pri;
    char * name;
    unsigned long count;
    unsigned long cumcount;
};
typedef struct Trans_descr * Trans_p;


static float tabnorm[14][2]= {
	{70,1.04},{75,1.15},{80,1.3},{85,1.44},{90,1.64},{91,1.70},{92,1.76},
	{93,1.82},{94,1.88},{95,1.96},{96,2.06},{97,2.18},{98,2.32},{99,2.58}
};

static float tabstud[30][14][2]={
	{                                              /*Grado di liberta'1*/
	{70,1.9626},{75,2.4142},{80,3.0777},{85,4.1653},{90,6.3138},
	{91,7.0264},{92,7.9158},{93,9.0579},{94,10.57},{95,12.7062},
	{96,15.8945},{97,21.2050},{98,31.8205},{99,63.6568}},
	{                                              /*Grado di liberta'2*/
	{70,1.3862},{75,1.6036},{80,1.8856},{85,2.2819},{90,2.9200},
	{91,3.1040},{92,3.3198},{93,3.5782},{94,3.8964},{95,4.3027},
	{96,4.8487},{97,5.6428},{98,6.9646},{99,9.9248}},
	{                                              /*Grado di liberta'3*/
	{70,1.2498},{75,1.4226},{80,1.6377},{85,1.9243},{90,2.3534},
	{91,2.4708},{92,2.6054},{93,2.7626},{94,2.9505},{95,3.1824},
	{96,3.4819},{97,3.8960},{98,4.5407},{99,5.8409}},
	{                                              /*Grado di liberta'4*/
	{70,1.1896},{75,1.3444},{80,1.5332},{85,1.7782},{90,2.1318},
	{91,2.2261},{92,2.3329},{93,2.4559},{94,2.6008},{95,2.7764},
	{96,2.9986},{97,3.2976},{98,3.7470},{99,4.6041}},
	{                                              /*Grado di liberta'5*/
	{70,1.1558},{75,1.3009},{80,1.4759},{85,1.6994},{90,2.0150},
	{91,2.0978},{92,2.1910},{93,2.2974},{94,2.4216},{95,2.5706},
	{96,2.7565},{97,3.0029},{98,3.3649},{99,4.0322}},
	{                                              /*Grado di liberta'6*/
	{70,1.1342},{75,1.2733},{80,1.4398},{85,1.6502},{90,1.9432},
	{91,2.0192},{92,2.1043},{93,2.2011},{94,2.3133},{95,2.4469},
	{96,2.6122},{97,2.8289},{98,3.1427},{99,3.7074}},
	{                                              /*Grado di liberta'7*/
	{70,1.1192},{75,1.2543},{80,1.4149},{85,1.6166},{90,1.8943},
	{91,1.9662},{92,2.0460},{93,2.1365},{94,2.2409},{95,2.3646},
	{96,2.5168},{97,2.7146},{98,2.9980},{99,3.4995}},
	{                                              /*Grado di liberta'8*/
	{70,1.1081},{75,1.2403},{80,1.3968},{85,1.5922},{90,1.8595},
	{91,1.9280},{92,2.0042},{93,2.0902},{94,2.1892},{95,2.3060},
	{96,2.4490},{97,2.6338},{98,2.8965},{99,3.3554}},
	{                                              /*Grado di liberta'9*/
	{70,1.0997},{75,1.2297},{80,1.3830},{85,1.5737},{90,1.8331},
	{91,1.8992},{92,1.9727},{93,2.0554},{94,2.1504},{95,2.2622},
	{96,2.3984},{97,2.5738},{98,2.8214},{99,3.2498}},
	{                                              /*Grado di liberta'10*/
	{70,1.0931},{75,1.2213},{80,1.3722},{85,1.5592},{90,1.8125},
	{91,1.8768},{92,1.9481},{93,2.0283},{94,2.1202},{95,2.2281},
	{96,2.3593},{97,2.5275},{98,2.7638},{99,3.1693}},
	{                                              /*Grado di liberta'11*/
	{70,1.0877},{75,1.2145},{80,1.3634},{85,1.5476},{90,1.7959},
	{91,1.8588},{92,1.9284},{93,2.0067},{94,2.0961},{95,2.2010},
	{96,2.3281},{97,2.4907},{98,2.7181},{99,3.1058}},
	{                                              /*Grado di liberta'12*/
	{70,1.0832},{75,1.2089},{80,1.3562},{85,1.5380},{90,1.7823},
	{91,1.8440},{92,1.9123},{93,1.9889},{94,2.0764},{95,2.1788},
	{96,2.3027},{97,2.4607},{98,2.6810},{99,3.0545}},
	{                                              /*Grado di liberta'13*/
	{70,1.0795},{75,1.2041},{80,1.3502},{85,1.5299},{90,1.7709},
	{91,1.8317},{92,1.8989},{93,1.9742},{94,2.0600},{95,2.1604},
	{96,2.2816},{97,2.4358},{98,2.6503},{99,3.0123}},
	{                                              /*Grado di liberta'14*/
	{70,1.0763},{75,1.2001},{80,1.3450},{85,1.5231},{90,1.7613},
	{91,1.8213},{92,1.8875},{93,1.9617},{94,2.0462},{95,2.1448},
	{96,2.2638},{97,2.4149},{98,2.6245},{99,2.9768}},
	{                                              /*Grado di liberta'15*/
	{70,1.0735},{75,1.1967},{80,1.3406},{85,1.5172},{90,1.7531},
	{91,1.8123},{92,1.8777},{93,1.9509},{94,2.0343},{95,2.1315},
	{96,2.2485},{97,2.3970},{98,2.6025},{99,2.9467}},
	{                                              /*Grado di liberta'16*/
	{70,1.0711},{75,1.1967},{80,1.3368},{85,1.5121},{90,1.7459},
	{91,1.8046},{92,1.8693},{93,1.9417},{94,2.0240},{95,2.1199},
	{96,2.2354},{97,2.3815},{98,2.5835},{99,2.9208}},
	{                                              /*Grado di liberta'17*/
	{70,1.0690},{75,1.1910},{80,1.3334},{85,1.5077},{90,1.7396},
	{91,1.7978},{92,1.8619},{93,1.9335},{94,2.0150},{95,2.1098},
	{96,2.2238},{97,2.3681},{98,2.5669},{99,2.8982}},
	{                                              /*Grado di liberta'18*/
	{70,1.0672},{75,1.1887},{80,1.3304},{85,1.5037},{90,1.7341},
	{91,1.7918},{92,1.8553},{93,1.9264},{94,2.0071},{95,2.1009},
	{96,2.2137},{97,2.3562},{98,2.5524},{99,2.8784}},
	{                                              /*Grado di liberta'19*/
	{70,1.0655},{75,1.1866},{80,1.3277},{85,1.5002},{90,1.7291},
	{91,1.7864},{92,1.8495},{93,1.9200},{94,2.0000},{95,2.0930},
	{96,2.2047},{97,2.3456},{98,2.5395},{99,2.8609}},
	{                                              /*Grado di liberta'20*/
	{70,1.0640},{75,1.1848},{80,1.3253},{85,1.4970},{90,1.7247},
	{91,1.7816},{92,1.8443},{93,1.9143},{94,1.9937},{95,2.0860},
	{96,2.1967},{97,2.3362},{98,2.5280},{99,2.8453}},
	{                                              /*Grado di liberta'21*/
	{70,1.0627},{75,1.1831},{80,1.3232},{85,1.4942},{90,1.7207},
	{91,1.7773},{92,1.8397},{93,1.9092},{94,1.9880},{95,2.0796},
	{96,2.1894},{97,2.3278},{98,2.5176},{99,2.8314}},
	{                                              /*Grado di liberta'22*/
	{70,1.0614},{75,1.1815},{80,1.3212},{85,1.4916},{90,1.7171},
	{91,1.7734},{92,1.8354},{93,1.9045},{94,1.9829},{95,2.0739},
	{96,2.1829},{97,2.3202},{98,2.5083},{99,2.8188}},
	{                                              /*Grado di liberta'23*/
	{70,1.0603},{75,1.1802},{80,1.3195},{85,1.4893},{90,1.7139},
	{91,1.7699},{92,1.8316},{93,1.9003},{94,1.9782},{95,2.0687},
	{96,2.1770},{97,2.3132},{98,2.4999},{99,2.8073}},
	{                                              /*Grado di liberta'24*/
	{70,1.0593},{75,1.1789},{80,1.3178},{85,1.4871},{90,1.7109},
	{91,1.7667},{92,1.8281},{93,1.8965},{94,1.9740},{95,2.0639},
	{96,2.1715},{97,2.3069},{98,2.4922},{99,2.7969}},
	{                                              /*Grado di liberta'25*/
	{70,1.0584},{75,1.1777},{80,1.3163},{85,1.4852},{90,1.7081},
	{91,1.7637},{92,1.8248},{93,1.8929},{94,1.9701},{95,2.0595},
	{96,2.1666},{97,2.3011},{98,2.4851},{99,2.7874}},
	{                                              /*Grado di liberta'26*/
	{70,1.0575},{75,1.1766},{80,1.3150},{85,1.4834},{90,1.7056},
	{91,1.7610},{92,1.8219},{93,1.8897},{94,1.9665},{95,2.0555},
	{96,2.1620},{97,2.2958},{98,2.4786},{99,2.7787}},
	{                                              /*Grado di liberta'27*/
	{70,1.0567},{75,1.1756},{80,1.3137},{85,1.4817},{90,1.7033},
	{91,1.7585},{92,1.8191},{93,1.8867},{94,1.9632},{95,2.0518},
	{96,2.1578},{97,2.2909},{98,2.4727},{99,2.7707}},
	{                                              /*Grado di liberta'28*/
	{70,1.0560},{75,1.1747},{80,1.3125},{85,1.4801},{90,1.7011},
	{91,1.7561},{92,1.8166},{93,1.8839},{94,1.9601},{95,2.0484},
	{96,2.1539},{97,2.2864},{98,2.4671},{99,2.7633}},
	{                                              /*Grado di liberta'29*/
	{70,1.0553},{75,1.1739},{80,1.3114},{85,1.4774},{90,1.6991},
	{91,1.7540},{92,1.8142},{93,1.8813},{94,1.9573},{95,2.0452},
	{96,2.1503},{97,2.2822},{98,2.4620},{99,2.7564}},
	{                                              /*Grado di liberta'30*/
	{70,1.0547},{75,1.1731},{80,1.3104},{85,1.4774},{90,1.6973},
	{91,1.7520},{92,1.8120},{93,1.8789},{94,1.9546},{95,2.0423},
	{96,2.1470},{97,2.2783},{98,2.4573},{99,2.7500}}
	 
};

extern char * ecalloc();


extern unsigned place_num, group_num, trans_num;
extern unsigned sub_num;
extern /*  PLACE ARRAY  */ Place_p p_list;
extern /* CODING BYTES */ unsigned no_cod;
extern struct Byte_descr * encoding;
extern /* DECODED MARKING */ unsigned short *DM;
extern unsigned char * cmarks, * c_m_p;
extern unsigned maxprob;
extern unsigned maxres;
extern Prob_p Prob;
extern Res_p Res;
extern double cur_time;
extern Trans_p t_list;
extern /* CONFIDENCE LEVEL */ int confidence;

